package resourcemgr

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"strings"
	"time"

	"github.com/mitchellh/mapstructure"

	"github.com/odpf/optimus/config"
	"github.com/odpf/optimus/models"
)

// ResourceManager is repository for external job spec
type ResourceManager interface {
	GetOptimusDependencies(context.Context, models.UnresolvedJobDependency) ([]models.OptimusDependency, error)
	GetExternalJobRuns(ctx context.Context, host, jobName, projectName string, startDate, endDate time.Time) ([]models.JobRun, error)
	GetHost() string
}

type optimusResourceManager struct {
	name   string
	config config.ResourceManagerConfigOptimus

	httpClient *http.Client
}

// NewOptimusResourceManager initializes job spec repository for Optimus neighbor
func NewOptimusResourceManager(resourceManagerConfig config.ResourceManager) (ResourceManager, error) {
	var conf config.ResourceManagerConfigOptimus
	if err := mapstructure.Decode(resourceManagerConfig.Config, &conf); err != nil {
		return nil, fmt.Errorf("error decoding resource manger config: %w", err)
	}
	if conf.Host == "" {
		return nil, errors.New("optimus resource manager host is empty")
	}
	return &optimusResourceManager{
		name:       resourceManagerConfig.Name,
		config:     conf,
		httpClient: http.DefaultClient,
	}, nil
}

func (o *optimusResourceManager) GetHost() string {
	return o.config.Host
}

func (o *optimusResourceManager) GetOptimusDependencies(ctx context.Context, unresolvedDependency models.UnresolvedJobDependency) ([]models.OptimusDependency, error) {
	if ctx == nil {
		return nil, errors.New("context is nil")
	}
	request, err := o.constructGetJobSpecificationsRequest(ctx, unresolvedDependency)
	if err != nil {
		return nil, fmt.Errorf("error encountered when constructing request: %w", err)
	}

	response, err := o.httpClient.Do(request)
	if err != nil {
		return nil, fmt.Errorf("error encountered when sending request: %w", err)
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("unexpected status response: %s", response.Status)
	}

	var jobSpecResponse getJobSpecificationsResponse
	decoder := json.NewDecoder(response.Body)
	if err := decoder.Decode(&jobSpecResponse); err != nil {
		return nil, fmt.Errorf("error decoding response: %w", err)
	}

	return o.toOptimusDependencies(jobSpecResponse.JobSpecificationResponses), nil
}

func (o *optimusResourceManager) constructGetJobSpecificationsRequest(ctx context.Context, unresolvedDependency models.UnresolvedJobDependency) (*http.Request, error) {
	var filters []string
	if unresolvedDependency.JobName != "" {
		filters = append(filters, fmt.Sprintf("job_name=%s", unresolvedDependency.JobName))
	}
	if unresolvedDependency.ProjectName != "" {
		filters = append(filters, fmt.Sprintf("project_name=%s", unresolvedDependency.ProjectName))
	}
	if unresolvedDependency.ResourceDestination != "" {
		filters = append(filters, fmt.Sprintf("resource_destination=%s", unresolvedDependency.ResourceDestination))
	}

	path := "/api/v1beta1/jobs"
	url := o.config.Host + path + "?" + strings.Join(filters, "&")

	request, err := http.NewRequestWithContext(ctx, "GET", url, http.NoBody)
	if err != nil {
		return nil, err
	}

	request.Header.Set("Accept", "application/json")
	for key, value := range o.config.Headers {
		request.Header.Set(key, value)
	}
	return request, nil
}

func (o *optimusResourceManager) toOptimusDependencies(responses []jobSpecificationResponse) []models.OptimusDependency {
	output := make([]models.OptimusDependency, len(responses))
	for i, r := range responses {
		output[i] = o.toOptimusDependency(r)
	}
	return output
}

func (o *optimusResourceManager) toOptimusDependency(response jobSpecificationResponse) models.OptimusDependency {
	return models.OptimusDependency{
		Name:          o.name,
		Host:          o.config.Host,
		Headers:       o.config.Headers,
		ProjectName:   response.ProjectName,
		NamespaceName: response.NamespaceName,
		JobName:       response.Job.Name,
		TaskName:      response.Job.TaskName,
	}
}

func (o *optimusResourceManager) GetExternalJobRuns(ctx context.Context, host, jobName, projectName string, startDate, endDate time.Time) ([]models.JobRun, error) {
	if ctx == nil {
		return nil, errors.New("context is nil")
	}
	request, err := o.constructGetJobRunRequest(ctx, host, jobName, projectName, startDate, endDate)
	if err != nil {
		return nil, fmt.Errorf("error encountered when constructing request: %w", err)
	}

	response, err := o.httpClient.Do(request)
	if err != nil {
		return nil, fmt.Errorf("error encountered when sending request: %w", err)
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("unexpected status response: %s", response.Status)
	}

	var jobRunResponse getJobRunResponse
	decoder := json.NewDecoder(response.Body)
	if err := decoder.Decode(&jobRunResponse); err != nil {
		return nil, fmt.Errorf("error decoding response: %w", err)
	}

	return toOptimusJobRuns(jobRunResponse.JobRunResponse)
}

func (o *optimusResourceManager) constructGetJobRunRequest(ctx context.Context, host, jobName, projectName string, startDate, endDate time.Time) (*http.Request, error) {
	var queryParams []string
	queryParams = append(queryParams, fmt.Sprintf("start_date=%s", startDate.Format(models.InstanceScheduledAtTimeLayout)))
	queryParams = append(queryParams, fmt.Sprintf("end_date=%s", endDate.Format(models.InstanceScheduledAtTimeLayout)))

	path := fmt.Sprintf("/api/v1beta1/project/%s/job/%s/run", projectName, jobName)
	url := host + path + "?" + strings.Join(queryParams, "&")

	request, err := http.NewRequestWithContext(ctx, "GET", url, http.NoBody)
	if err != nil {
		return nil, err
	}

	request.Header.Set("Accept", "application/json")
	for key, value := range o.config.Headers {
		request.Header.Set(key, value)
	}
	return request, nil
}

func toOptimusJobRuns(responses []jobRunResponse) ([]models.JobRun, error) {
	output := make([]models.JobRun, len(responses))
	for i, r := range responses {
		jobRun, err := toOptimusJobRun(r)
		if err != nil {
			return output, err
		}
		output[i] = jobRun
	}
	return output, nil
}

func toOptimusJobRun(response jobRunResponse) (models.JobRun, error) {
	parsedScheduleTime, err := time.Parse(models.InstanceScheduledAtTimeLayout, response.ScheduledAt)

	if err != nil {
		return models.JobRun{}, err
	}
	return models.JobRun{
		Status:      models.JobRunState(response.State),
		ScheduledAt: parsedScheduleTime,
	}, err
}
